﻿using Syncfusion.Data;
using Syncfusion.UI.Xaml.Grid;
using Syncfusion.UI.Xaml.Grid.Helpers;
using Syncfusion.UI.Xaml.ScrollAxis;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;

namespace SfDataGridDemo
{
    public class AutoScrollerExt : AutoScroller
    {
        public Point MousePoint
        {
            get { return this.MouseMovePosition; }
            set { this.MouseMovePosition = value; }
        }
        public AutoScrollerExt()
            : base()
        {
            this.MousePoint = new Point(0, 0);
        }

    }
    public class GridSelectionControllerExt<T> : GridSelectionController
    {
        #region Fields
        bool MousePressed = false;
        public Popup popup;
        public Popup upArrowIndicator;
        public Popup downArrowIndicator;
        public DraggablePopupContentControl draggablePopupContentControl = null;

        NodeEntry _draggingRecordEntry = null;
        private bool _isDragging;

        int previousRowIndex = -1;
        #endregion

        #region Properties
        public bool IsDragging
        {
            get { return _isDragging; }
            set { _isDragging = value; }
        }

        public VisualContainer visualcontainer = null;

        #endregion

        #region ctr

        public GridSelectionControllerExt(SfDataGrid sfdatagrid)
            : base(sfdatagrid)
        {
            this.DataGrid.Loaded += DataGrid_Loaded;
        }

        public void DataGrid_Loaded(object sender, RoutedEventArgs e)
        {
            visualcontainer = this.DataGrid.GetVisualContainer();
            this.DataGrid.AutoScroller = new AutoScrollerExt();
            UpdateAutoScroller();
        }

        #endregion

        public event EventHandler CanExecuteChanged;
        public string ClassName;

        protected override void ProcessKeyDown(KeyEventArgs args)
        {
            base.ProcessKeyDown(args);
        }

        #region Overrides
        protected override void ProcessOnDoubleTapped(MouseButtonEventArgs e, RowColumnIndex currentRowColumnIndex)
        {
            base.ProcessOnDoubleTapped(e, currentRowColumnIndex);
            MousePressed = false;
        }
        #endregion

        #region Overrides
        protected override void ProcessPointerPressed(MouseButtonEventArgs args, RowColumnIndex rowColumnIndex)
        {
            if (!args.ChangedButton.ToString().Equals("Right"))
            {
                MousePressed = true;
            }
            base.ProcessPointerPressed(args, rowColumnIndex);
        }
        #endregion

        #region Overrides
        protected override void ProcessPointerReleased(MouseButtonEventArgs args, RowColumnIndex rowColumnIndex)
        {
            base.ProcessPointerReleased(args, rowColumnIndex);
            MousePressed = false;
        }
        #endregion

        #region Overrides
        protected override void ProcessPointerMoved(MouseEventArgs args, RowColumnIndex rowColumnIndex)
        {
            if (!MousePressed || _isDragging)
                return;
            if (rowColumnIndex.RowIndex >= this.DataGrid.ResolveStartIndexBasedOnPosition())
            { 
                base.ProcessPointerMoved(args, rowColumnIndex);
                _draggingRecordEntry = DataGrid.GetNodeEntry(rowColumnIndex.RowIndex);
                if (_draggingRecordEntry.IsGroups)
                    this.DataGrid.View.TopLevelGroup.ResetCache = true;
                if (!_draggingRecordEntry.IsRecords)
                {
                    _draggingRecordEntry = null;
                    return;
                }
                _isDragging = true;
                InitializePopupControl();

                // Get the exact position where this mouse pointer is pressed
                var p = args.GetPosition(this.visualcontainer);

                var gridRect = this.GetControlRect(this.DataGrid);
                // Set the Horizontal and Vertical offset for popup screen
                popup.HorizontalOffset = p.X;
                popup.VerticalOffset = p.Y;
                //Show the Drag indicator            
                this.ShowDragIndication(gridRect, p);
                popup.IsOpen = true;
                this.SuspendAutoScrolling = true;
                draggablePopupContentControl.CaptureMouse();
            }
        }
        #endregion

        #region Events of DraggablePopupContentControl

        private void OnDrop(object sender, MouseButtonEventArgs e)
        {
            MousePressed = false;
            if (!this.IsDragging || this.popup == null)
                return;

            if (_draggingRecordEntry.Parent != null)
                DragAndDropWhenGroup(e);
            else
                DragAndDropWithRecord(e);

            _isDragging = false;
            popup.IsOpen = false;
            upArrowIndicator.IsOpen = false;
            downArrowIndicator.IsOpen = false;
            this.DataGrid.AutoScroller.AutoScrolling = AutoScrollOrientation.None;
        }

        private void DragOver(object sender, MouseEventArgs e)
        {
            if (!IsDragging || this.popup == null || !(sender is DraggablePopupContentControl))
                return;
            this.draggablePopupContentControl.CaptureMouse();
            var p = e.GetPosition(this.visualcontainer);

            if (!IsPointInsideSfDataGrid(p))
            {
                var pointToContentControl = draggablePopupContentControl.PointToScreen(e.GetPosition(draggablePopupContentControl));
                popup.HorizontalOffset = p.X;
                popup.VerticalOffset = p.Y;
                upArrowIndicator.IsOpen = downArrowIndicator.IsOpen = false;
                var visual = this.DataGrid.GetVisualContainer();
                this.SuspendAutoScrolling = true;
                this.DataGrid.AutoScroller.AutoScrolling = AutoScrollOrientation.Vertical;
            }
            else
            {
                // Set the Horizontal and Vertical offset for popup screen
                popup.HorizontalOffset = p.X;
                popup.VerticalOffset = p.Y;
                this.ShowDragIndication(this.GetControlRect(this.DataGrid), p);
                popup.IsOpen = true;
            }
            (this.DataGrid.AutoScroller as AutoScrollerExt).MousePoint = e.GetPosition(this.DataGrid);
            e.Handled = true;
        }

        #endregion

        #region Methods

        void UpdateAutoScroller()
        {
            this.DataGrid.AutoScroller.VisualContainer = this.visualcontainer;
            this.DataGrid.AutoScroller.AutoScrollBounds = this.visualcontainer.GetClipRect(ScrollAxisRegion.Header, ScrollAxisRegion.Footer);
            this.DataGrid.AutoScroller.IntervalTime = new TimeSpan(0, 0, 0, 0, 40);
            this.DataGrid.AutoScroller.InsideScrollMargin = new Size(0, 0);
        }

        private void InitializePopupControl()
        {
            if (this.popup != null)
            {
                draggablePopupContentControl.DataContext = _draggingRecordEntry;
                return;
            }

            //Initialize Popup, UpArrow, DownArrow and DraggablePopupContentControl
            popup = new Popup();
            upArrowIndicator = new Popup();
            downArrowIndicator = new Popup();
            draggablePopupContentControl = new DraggablePopupContentControl(this.DataGrid) { DataContext = _draggingRecordEntry };

            //Set width and Height of popup
            popup.Height = 110;
            popup.Width = 250;

            //Set the placement target for popup ,UpArrowIndicator and DownArrowIndicator
            popup.Placement = PlacementMode.Relative;
            popup.PlacementTarget = this.visualcontainer;
            popup.AllowsTransparency = true;

            upArrowIndicator.Placement = PlacementMode.Relative;
            upArrowIndicator.PlacementTarget = this.visualcontainer;
            upArrowIndicator.AllowsTransparency = true;

            downArrowIndicator.Placement = PlacementMode.Relative;
            downArrowIndicator.PlacementTarget = this.visualcontainer;
            downArrowIndicator.AllowsTransparency = true;

            popup.Child = draggablePopupContentControl;
            upArrowIndicator.Child = new UpIndicatorContentControl();
            downArrowIndicator.Child = new DownIndicatorContentControl();

            draggablePopupContentControl.PreviewMouseMove += DragOver;
            draggablePopupContentControl.PreviewMouseLeftButtonUp += OnDrop;
            draggablePopupContentControl.PreviewMouseDown += DraggablePopupContentControl_PreviewMouseDown;
        }

        private void DraggablePopupContentControl_PreviewMouseDown(object sender, MouseButtonEventArgs e)
        {
            _isDragging = false;
        }

        private void ShowDragIndication(Rect gridRect, Point mousePoint)
        {
            var currentRowColumnIndex = visualcontainer.PointToCellRowColumnIndex(Mouse.GetPosition(visualcontainer));

            if (currentRowColumnIndex.IsEmpty || currentRowColumnIndex.RowIndex < this.DataGrid.ResolveStartIndexBasedOnPosition())
                return;

            int currentRecordIndex = this.DataGrid.ResolveToRecordIndex(currentRowColumnIndex.RowIndex);
            if (currentRecordIndex == -1)
                return;

            var rowGenerator = this.DataGrid.GetRowGenerator();
            var nodeentry = DataGrid.GetNodeEntry(currentRowColumnIndex.RowIndex);

            if (!nodeentry.IsRecords)
                return;
            var recordUnderMouse = nodeentry as RecordEntry;
            var dataRow = rowGenerator.Items.FirstOrDefault(row => (row.RowType != RowType.HeaderRow && (row.RowData  == recordUnderMouse.Data))) as DataRowBase;
            //Get the whole row element  for DataRow
            var wholeRowELement = (dataRow as IElement).Element as VirtualizingCellsControl;
            //Get the mouse point position on the VirtualizingCellsControl
            var mousePointOnVirtualizingCellsControl = Mouse.GetPosition(wholeRowELement);
            //Set the Horizontal and Vertical Offset for UpAprrowIndicator and DownArrowIndicator
            upArrowIndicator.HorizontalOffset = DataGrid.RowHeaderWidth;
            downArrowIndicator.HorizontalOffset = DataGrid.RowHeaderWidth;

            //Check whether the mouse point is on the top or end of the Row element 
            if (currentRecordIndex != previousRowIndex || (currentRecordIndex == previousRowIndex && mousePointOnVirtualizingCellsControl.Y < this.DataGrid.RowHeight / 2))
            {
                previousRowIndex = currentRecordIndex;

                upArrowIndicator.VerticalOffset = mousePoint.Y - mousePointOnVirtualizingCellsControl.Y;
                downArrowIndicator.VerticalOffset = mousePoint.Y - (mousePointOnVirtualizingCellsControl.Y + 22);
            }
            else if (currentRecordIndex == previousRowIndex && mousePointOnVirtualizingCellsControl.Y > this.DataGrid.RowHeight / 2)
            {
                upArrowIndicator.VerticalOffset = mousePoint.Y + (this.DataGrid.RowHeight - mousePointOnVirtualizingCellsControl.Y);
                downArrowIndicator.VerticalOffset = mousePoint.Y + (this.DataGrid.RowHeight - mousePointOnVirtualizingCellsControl.Y) - 22;
            }

            upArrowIndicator.IsOpen = true;
            downArrowIndicator.IsOpen = true;
        }

        private void DragAndDropWhenGroup(MouseButtonEventArgs e)
        {
            var point = e.GetPosition(visualcontainer);
            var rowColumnIndex = visualcontainer.PointToCellRowColumnIndex(point);
            bool isSameGroup = false;

            if (!_isDragging)
                return;

            // Get the droppingRecordEntry using RowIndex            
            var droppingRecordEntry = DataGrid.GetNodeEntry(rowColumnIndex.RowIndex);

            if (droppingRecordEntry == null)
                return;

            if (!droppingRecordEntry.IsRecords)
                return;

            if (droppingRecordEntry.Parent == _draggingRecordEntry.Parent)
                isSameGroup = true;

            if (_draggingRecordEntry == droppingRecordEntry)
            {
                return;
            }
            var draggingGroup = _draggingRecordEntry.Parent as Group;
            draggingGroup.RemoveRecord(_draggingRecordEntry as RecordEntry, true);
            draggingGroup.SetDirty();
            int indexOfDroppingRecord = (droppingRecordEntry.Parent as Group).Records.IndexOf(droppingRecordEntry as RecordEntry);
            if (!isSameGroup)
            {
                _draggingRecordEntry.Parent = null;
            }
            _draggingRecordEntry.Parent = droppingRecordEntry.Parent;
            (droppingRecordEntry.Parent as Group).InsertRecord(indexOfDroppingRecord, _draggingRecordEntry as RecordEntry, true);
            this.DataGrid.View.TopLevelGroup.ResetCache = true;
            this.DataGrid.View.TopLevelGroup.SetDirty();
            this.DataGrid.GetGridModel().RefreshView(true);
            this.DataGrid.SelectedItem = ((_draggingRecordEntry.Parent as Group).Records[indexOfDroppingRecord] as RecordEntry).Data;
        }

        private void DragAndDropWithRecord(MouseButtonEventArgs e)
        {
            var point = e.GetPosition(visualcontainer);
            var upArrowPoint = new Point(upArrowIndicator.HorizontalOffset, upArrowIndicator.VerticalOffset - 22);
            var downArrowPoint = new Point(downArrowIndicator.HorizontalOffset, downArrowIndicator.VerticalOffset + 22);
            var rowColumnIndex = visualcontainer.PointToCellRowColumnIndex(point);
            var currentRowIndex = DataGrid.ResolveToRowIndex(_draggingRecordEntry);

            point = rowColumnIndex.RowIndex > currentRowIndex ? upArrowPoint : downArrowPoint;

            rowColumnIndex = visualcontainer.PointToCellRowColumnIndex(point);
            if (!_isDragging)
                return;
            if (!(this.DataGrid is DetailsViewDataGrid))
            {
                var vm = DataGrid.ItemsSource as ObservableCollection<T>;
                if (!rowColumnIndex.IsEmpty)
                {
                    // Get the RecordIndex using RowIndex
                    var index = DataGrid.ResolveToRecordIndex(rowColumnIndex.RowIndex);
                    
					if (index != -1)
					{
						var dragrecord = (T)(_draggingRecordEntry as RecordEntry).Data;
						DataGrid.SelectedItem = vm[index];
					}
					else
					{
						var dragrecord = (T)(_draggingRecordEntry as RecordEntry).Data;
						// Remove the Selected Row Data from Dragging SfDataGrid 
						vm.Remove(dragrecord);
						// Insert the Selected row data to specified index position of dropping SfDataGrid
						vm.Insert(index + 1, dragrecord);
						// Selected Item is reset with new dropped row data
						DataGrid.SelectedItem = (_draggingRecordEntry as RecordEntry).Data;
						this.MoveCurrentCell(new RowColumnIndex(rowColumnIndex.RowIndex + 1, 1));
					}
                }
                else
                {
                    return;
                }
            }
        }

        #endregion

        #region HelperMethods

        internal Rect GetControlRect(FrameworkElement control)
        {
            var locationfromWindow = control.TranslatePoint(new Point(0, 0), control);
            var locationfromScreen = locationfromWindow;
            return new Rect((locationfromScreen.X - locationfromWindow.X),
                                (locationfromScreen.Y - locationfromWindow.Y),
                                control.ActualWidth, control.ActualHeight);
        }

        private bool IsPointInsideSfDataGrid(Point point)
        {
            return this.GetControlRect(this.DataGrid).Contains(point);
        }

        protected override void Dispose(bool isDisposing)
        {
            base.Dispose();
            if (draggablePopupContentControl == null) return;
            draggablePopupContentControl.PreviewMouseMove -= DragOver;
            draggablePopupContentControl.PreviewMouseLeftButtonUp -= OnDrop;
            draggablePopupContentControl.Dispose();
        }
        #endregion
    }

    public class DraggablePopupContentControl : ContentControl, IDisposable
    {
        internal SfDataGrid dataGrid;

        public DraggablePopupContentControl(SfDataGrid DataGrid)
        {
            base.DefaultStyleKey = typeof(DraggablePopupContentControl);
            dataGrid = DataGrid;
            this.Loaded += DraggablePopupContentControl_Loaded;
        }

        void DraggablePopupContentControl_Loaded(object sender, RoutedEventArgs e)
        {
			var stackPanel = new StackPanel() { Orientation = Orientation.Vertical, Width = 100, Margin = new Thickness(4) };
            this.Content = stackPanel;
        }

        public void Dispose()
        {
            this.dataGrid = null;
            this.Loaded -= DraggablePopupContentControl_Loaded;
        }
    }
}